using System;
using System.Linq;
using UnityEngine;

public class CircleGenerator : ShapeGenerator
{
    #region InspectorVariables
    [Header("Circle values")]
    [Range(0,100)] public float radius = 1f;

    [Header("Technical")]
    [Range(3,1000)] public int sideCount = 500;
    #endregion InspectorVariables

    #region Implementation
    protected override (Vector3[] vertices, int[] triangles) GenerateTopology()
        => CreateCircle(radius, sideCount);

    /// <summary>
    /// Creates circle vertecies and triangles around the midPoint at (0,0,0)
    /// </summary>
    /// <param name="points"></param>
    /// <param name="invert"></param>
    /// <returns></returns>
    public static (Vector3[], int[]) CreateCircle(float radius, int sideCount, Vector3 offset, bool invert = false)
    {
        Vector3[] edge = GetCirclePoints(radius, sideCount);
        Vector3[] vertices =
            new Vector3[] { offset }
            .AppendRange(edge)
            .Append(edge[0])
            .ToArray();

        return CreatePlane(vertices, invert);
    }

    public static (Vector3[], int[]) CreateCircle(float radius, int sideCount, bool invert = false)
        => CreateCircle(radius, sideCount, Vector3.zero, invert);
    #endregion Implementation
}
