using UnityEngine;
using UnityEngine.EventSystems;
using static UnityEngine.EventSystems.PointerEventData;

[RequireComponent(typeof(FactWrapper), typeof(RectTransform), typeof(DragHandling))]
public class OpenFactExplorer : MonoBehaviour, IPointerClickHandler
{
    #region Variables
    public GameObject factExplorerPrefab;

    private static Transform factExplorer;
    private float pressTime = 0f;
    private const float LONG_PRESS_DURATION = 0.5f;
    private bool beganTouch = false;

    private DragHandling CachedDragHandling;
    private FactWrapper CachedFactWrapper;
    #endregion Variables

    #region UnityMethods
    public void OnPointerClick(PointerEventData eventData)
    { // Handles PC clicks
        if (eventData.dragging)
            return;

        if (eventData.button == InputButton.Right)
            DoOpenFactExplorer();
        else
        if (eventData.button == InputButton.Left
         )//&& eventData.clickCount >= 2)
            DoSetActive();
    }

    private void Awake()
    {
        CachedDragHandling = GetComponent<DragHandling>();
        CachedFactWrapper = GetComponent<FactWrapper>();
    }

    private void Update()
    {
        // open FactExplorer on press on fact longer than LONG_PRESS_DURATION
        HandleTouches();
    }
    #endregion UnityMethods

    #region Implementation
    private void HandleTouches()
    {
        if (Input.touchCount != 1 || CachedDragHandling.dragged)
        {
            pressTime = 0;
            return;
        }

        var touch = Input.GetTouch(0);
        if (!RectTransformUtility.RectangleContainsScreenPoint((RectTransform)transform, touch.position))
        {
            pressTime = 0;
            return;
        }

        switch (touch.phase)
        {
            case TouchPhase.Began:
                beganTouch = true;
                pressTime = 0;
                break;

            case TouchPhase.Ended:
                if (beganTouch
                 && pressTime < LONG_PRESS_DURATION)
                    DoSetActive();
                goto case TouchPhase.Canceled;

            case TouchPhase.Moved:
            case TouchPhase.Canceled:
                beganTouch = false;
                pressTime = 0;
                break;

            case TouchPhase.Stationary:
                pressTime += Time.deltaTime;
                if (pressTime >= LONG_PRESS_DURATION)
                {
                    DoOpenFactExplorer();
                    goto case TouchPhase.Canceled;
                }
                break;
        }
    }

    private void DoOpenFactExplorer()
    {
        if (factExplorer != null
         && factExplorer.GetComponent<FactExplorer>().MainFact.Id == CachedFactWrapper.URI)
        {
            Destroy(factExplorer.gameObject);
            return;
        }

        if (factExplorer != null)
            Destroy(factExplorer.gameObject);

        var parent = transform.GetComponentInParent<Canvas>().transform;

        factExplorer = Instantiate(factExplorerPrefab.transform, Input.mousePosition, Quaternion.identity, parent);
        factExplorer.GetComponent<FactExplorer>().Initialize(CachedFactWrapper.Fact, transform.position);
    }

    private bool DoSetActive()
        => ScrollDetails.Instance.SetNextEmptyTo(CachedFactWrapper as FactObjectUI);
    #endregion Implementation
}

