using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public static class GameObjectExtensions
{
    public static bool IsRoot(this GameObject root)
        => root.transform.parent == null;

    public static void SetActiveByTagRecursive(this GameObject root, string tag, bool enable)
        => root.ForAllChildren(child => {
            if (child.CompareTag(tag))
                child.SetActive(enable);
            else
                SetActiveByTagRecursive(child, tag, enable);
        });

    public static void DestroyAllChildren(this GameObject root)
        => root.ForAllChildren(child => GameObject.Destroy(child));

    public static void SetActiveAllChildren(this GameObject root, bool active)
        => root.ForAllChildren(child => child.SetActive(active));

    public static void ForAllChildren(this GameObject root, Action<GameObject> func_on_child)
    {
        foreach (Transform transform in root.transform)
            func_on_child(transform.gameObject);
    }

    public static List<GameObject> GetDirectChildren(this GameObject root)
    {
        List<GameObject> ret = new (capacity: root.transform.childCount);
        foreach (Transform transform in root.transform)
            ret.Add(transform.gameObject);

        return ret;
    }

    public static GameObject GetNthChild(this GameObject root, IEnumerable<int> pos)
    {
        GameObject ret = root;
        foreach (var i in pos)
            ret = ret.transform.GetChild(i).gameObject;

        return ret;
    }
}
