using System;
using System.Collections.Generic;
using System.Linq;

public static class TupleExtensions
{
    public static readonly Type[] GenericTupleTypes = new Type[]
{
    typeof(Tuple),
    typeof(Tuple<>),
    typeof(Tuple<,>),
    typeof(Tuple<,,>),
    typeof(Tuple<,,,>),
    typeof(Tuple<,,,,>),
    typeof(Tuple<,,,,,>),
    typeof(Tuple<,,,,,,>),
    typeof(Tuple<,,,,,,,>)
};

    private static readonly IReadOnlyDictionary<Type, int> GenericFuncSignature =
        GenericTupleTypes.ToDictionary(t => t, t => t.IsGenericType ? t.GetGenericArguments().Count() : 0);

    public static Type GetGenericTupleType(int argumentsCount)
        => GenericTupleTypes[argumentsCount];

    public static bool IsTupleType(Type type, out int signatureCount)
        => GenericFuncSignature.TryGetValue(
                type.IsGenericType ? type.GetGenericTypeDefinition() : type,
                out signatureCount
        );

    public static Type CreateTupleType(Type[] signature)
        => GetGenericTupleType(signature.Length)
            .MakeGenericType(signature);
}